package se.su.dsv.scipro.daisyExternal.http;

import brave.http.HttpTracing;
import brave.jaxrs2.TracingFeature;
import org.glassfish.jersey.client.authentication.HttpAuthenticationFeature;
import scala.Option;
import se.su.dsv.scipro.io.dto.*;

import javax.inject.Inject;
import javax.inject.Named;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.GenericType;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.io.InputStream;
import java.util.List;
import java.util.Optional;
import java.util.Set;

import static javax.ws.rs.client.Entity.json;
import static javax.ws.rs.client.Entity.xml;

public class DaisyAPIImpl implements DaisyAPI {

    private static final String STUDENT = "student";
    private static final String COURSE = "course";
    private static final String CONTRIBUTOR = "contributor";
    private static final String THESES = "theses";
    private static final String SUBUNITS = "subunits";
    private static final String USERNAMES = "usernames";
    private static final String RESEARCH_AREAS = "researchAreas";
    private static final String SUPERVISORS = "supervisors";
    private static final String PROGRAM = "program";
    private static final String THESIS = "thesis";
    private static final String ORGUNIT = "orgunit";
    private static final String PERSON = "person";
    private static final String USERNAME = "username";
    
    private final String baseUrl;
    private final String user;
    private final String password;
    private final HttpTracing httpTracing;

    @Inject
    public DaisyAPIImpl(
            @Named("url") final String baseUrl,
            @Named("user") final String user,
            @Named("password") final String password,
            final HttpTracing httpTracing)
    {
        this.baseUrl = baseUrl;
        this.user = user;
        this.password = password;
        this.httpTracing = httpTracing;
    }

    @Override
    public Set<ProjectParticipant> getContributors(Long projectId) {
        return thesis()
                .path(String.valueOf(projectId))
                .path(CONTRIBUTOR)
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get(new GenericType<Set<ProjectParticipant>>() {
                });
    }


    @Override
    public Set<StudentProjectParticipant> getAuthors(Long projectId) {
        return thesis()
                .path(String.valueOf(projectId))
                .path(STUDENT)
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get(new GenericType<Set<StudentProjectParticipant>>() {
                });
    }

    @Override
    public Set<Thesis> getProjects(Long personId) {
        return person()
                .path(String.valueOf(personId))
                .path(THESES)
                .request(MediaType.APPLICATION_XML_TYPE)
                .get(new GenericType<Set<Thesis>>() {
                });
    }

    @Override
    public Set<Unit> getSubUnits(Long unitId) {
        return units()
                .path(String.valueOf(unitId))
                .path(SUBUNITS)
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get(new GenericType<Set<Unit>>() {
                });
    }


    @Override
    public Set<UserName> getUsernames(Long personId) {
        return person()
                .path(String.valueOf(personId))
                .path(USERNAMES)
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get(new GenericType<Set<UserName>>() {
                });
    }

    @Override
    public List<ResearchArea> getResearchAreas(Long personId) {
        return person()
                .path(String.valueOf(personId))
                .path(RESEARCH_AREAS)
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get(new GenericType<List<ResearchArea>>() {
                });
    }

    @Override
    public List<Person> getSupervisors(Long unitId) {
        return units()
                .path(String.valueOf(unitId))
                .path(SUPERVISORS)
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get(new GenericType<List<Person>>() {
                });
    }

    @Override
    public Option<Person> findPersonById(Long id) {
        Response response = person()
                .path(String.valueOf(id))
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get(Response.class);
        return response.getStatus() == 200
                ? Option.apply(response.readEntity(Person.class))
                : Option.<Person>empty();
    }

    @Override
    public Optional<Person> findByUsername(String userName) {
        Response response = person()
                .path(USERNAME)
                .path(userName)
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get(Response.class);
        return response.getStatus() == 200
                ? Optional.of(response.readEntity(Person.class))
                : Optional.<Person>empty();
    }

    @Override
    public Response addContributor(Long projectIdentifier, ProjectParticipant contributor) {
        return thesis()
                .path(String.valueOf(projectIdentifier))
                .path(CONTRIBUTOR)
                .request(MediaType.APPLICATION_JSON_TYPE)
                .post(json(contributor));
    }

    @Override
    public Response deleteContributor(Long projectId, Long personId) {
        return thesis()
                .path(String.valueOf(projectId))
                .path(CONTRIBUTOR)
                .path(String.valueOf(personId))
                .request(MediaType.APPLICATION_JSON_TYPE)
                .delete();
    }

    @Override
    public Response createProject(Thesis project) {
        return thesis()
                .request(MediaType.APPLICATION_JSON_TYPE)
                .post(json(project));
    }

    @Override
    public Response addAuthor(Long projectId, StudentProjectParticipant author) {
        return thesis()
                .path(String.valueOf(projectId))
                .path(STUDENT)
                .request(MediaType.APPLICATION_XML_TYPE)
                .post(xml(author));
    }

    @Override
    public Response deleteProject(Long projectId) {
        return thesis()
                .path(String.valueOf(projectId))
                .request(MediaType.APPLICATION_JSON_TYPE)
                .delete();
    }

    @Override
    public Response getStudent(Long id) {
        return student()
                .path(String.valueOf(id))
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get();
    }

    @Override
    public Program getProgram(Long id) {
        return program()
                .path(String.valueOf(id))
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get(Program.class);
    }

    @Override
    public se.su.dsv.scipro.util.Option<Thesis> getThesis(Long projectIdentifier) {
        Response response = thesis()
                .path(String.valueOf(projectIdentifier))
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get(Response.class);
        return response.getStatus() == 200
                ? se.su.dsv.scipro.util.Option.of(response.readEntity(Thesis.class))
                : se.su.dsv.scipro.util.Option.empty();
     }

    @Override
    public void updateThesis(Thesis thesis) {
        thesis()
                .path(String.valueOf(thesis.getId()))
                .request()
                .put(xml(thesis));
    }

    @Override
    public void deleteAuthor(Long projectId, Long authorId) {
        thesis()
                .path(String.valueOf(projectId))
                .path(STUDENT)
                .path(String.valueOf(authorId))
                .request()
                .delete();
    }

    @Override
    public Option<String> getGradeForStudentInCourse(Long authorId, Long courseId) {
        Response response = course()
                .path(String.valueOf(courseId))
                .path(STUDENT)
                .path(String.valueOf(authorId))
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get();
        return response.getStatus() == 200
                ? Option.apply(response.readEntity(String.class))
                : Option.<String>empty();
    }

    @Override
    public List<Person> findByPersonnummer(final String personnummer) {
        return person()
                .queryParam("personnummer", personnummer)
                .request(MediaType.APPLICATION_JSON_TYPE)
                .get(new GenericType<List<Person>>(){});
    }

    @Override
    public PhotoResult getPhoto(final Long personId, final long requesterId, final boolean alwaysShow) {
        final Response response = person()
                .path(String.valueOf(personId))
                .path("photo")
                .queryParam("show", alwaysShow)
                .request()
                .header("X-User-Id", requesterId)
                .get();
        switch (response.getStatus()) {
            case 200:
                return new PhotoResult.Found(response.readEntity(InputStream.class));
            case 403:
                return new PhotoResult.Forbidden();
            case 404:
            case 503:
                return new PhotoResult.Missing();
            default:
                throw new RuntimeException("Unknown response from profile service: " + response);
        }
    }

    @Override
    public se.su.dsv.scipro.util.Option<ThesisPublication> getPublication(final long projectId) {
        final Response response = thesis()
                .path(Long.toString(projectId))
                .path("publication")
                .request(MediaType.APPLICATION_XML_TYPE)
                .get();
        switch (response.getStatus()) {
            case 200:
                return se.su.dsv.scipro.util.Option.of(response.readEntity(ThesisPublication.class));
            case 204:
            case 404:
                return se.su.dsv.scipro.util.Option.empty();
            default:
                throw new RuntimeException("Unknown response from thesis publication: " + response);
        }
    }

    @Override
    public boolean sendPublication(final long projectId, final ThesisPublication publication) {
        final Response response = thesis()
                .path(Long.toString(projectId))
                .path("publication")
                .request(MediaType.APPLICATION_XML_TYPE)
                .put(Entity.xml(publication));
        return response.getStatus() == 200;
    }

    @Override
    public boolean sendPublicationFile(final long projectId, final String filename, final long length, final InputStream data) {
        final Response response = thesis()
                .path(Long.toString(projectId))
                .path("publication")
                .path("file")
                .request(MediaType.APPLICATION_XML_TYPE)
                .header("X-Filename", filename)
                .put(Entity.entity(data, "application/pdf"));
        return response.getStatus() == 200;
    }

    private WebTarget program() {
        return target()
                .path(PROGRAM);
    }

    private WebTarget student() {
        return target()
                .path(STUDENT);
    }

    private WebTarget thesis() {
        return target()
                .path(THESIS);
    }

    private WebTarget units() {
        return target()
                .path(ORGUNIT);
    }

    private WebTarget person() {
        return target()
                .path(PERSON);
    }

    private WebTarget course() {
        return target()
                .path(COURSE);
    }

    private WebTarget target() {
        return ClientBuilder.newClient()
                .register(HttpAuthenticationFeature.basic(user, password))
                .register(TracingFeature.create(httpTracing))
                .target(baseUrl);
    }
}
